------------------------------------------------------------------------------
--                                                                          --
--                 ASIS-for-GNAT IMPLEMENTATION COMPONENTS                  --
--                                                                          --
--                           I N S T A L L _ A S I S                        --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--           Copyright (c) 2003, Free Software Foundation, Inc.             --
--                                                                          --
-- ASIS-for-GNAT is free software; you can redistribute it and/or modify it --
-- under terms of the  GNU General Public License  as published by the Free --
-- Software Foundation;  either version 2,  or  (at your option)  any later --
-- version. ASIS-for-GNAT is distributed  in the hope  that it will be use- --
-- ful, but WITHOUT ANY WARRANTY; without even the implied warranty of MER- --
-- CHANTABILITY or  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General  --
-- Public License for more details. You should have received a copy of the  --
-- GNU General Public License  distributed with ASIS-for-GNAT; see file     --
-- COPYING. If not, write to the Free Software Foundation,  59 Temple Place --
-- - Suite 330,  Boston, MA 02111-1307, USA.                                --
--                                                                          --
--                                                                          --
--                                                                          --
--                                                                          --
--                                                                          --
--                                                                          --
--                                                                          --
--                                                                          --
-- ASIS-for-GNAT was originally developed  by the ASIS-for-GNAT team at the --
-- Software  Engineering  Laboratory  of  the Swiss  Federal  Institute  of --
-- Technology (LGL-EPFL) in Lausanne,  Switzerland, in cooperation with the --
-- Scientific  Research  Computer  Center of  Moscow State University (SRCC --
-- MSU), Russia,  with funding partially provided  by grants from the Swiss --
-- National  Science  Foundation  and  the  Swiss  Academy  of  Engineering --
-- Sciences.  ASIS-for-GNAT is now maintained by  Ada Core Technologies Inc --
-- (http://www.gnat.com).                                                   --
--                                                                          --
------------------------------------------------------------------------------
--  The only purpose of this ASIS "application" is to enforce
--  the compilation of all the ASIS implementation and GNAT components
--  needed for any real ASIS application for creating the executable

with Asis;

with Asis.Ada_Environments;
with Asis.Ada_Environments.Containers;
with Asis.Clauses;
with Asis.Compilation_Units;
with Asis.Compilation_Units.Times;
with Asis.Compilation_Units.Relations;
with Asis.Data_Decomposition;
with Asis.Declarations;
with Asis.Definitions;
with Asis.Elements;
with Asis.Expressions;
with Asis.Ids;
with Asis.Implementation;
with Asis.Implementation.Permissions;
with Asis.Statements;
with Asis.Text;
with Asis.Iterator;

with Asis.Extensions;
with Asis.Extensions.Flat_Kinds;
with Asis.Extensions.Iterator;
with Asis.Data_Decomposition.Extensions;

with Text_IO; use Text_IO;

procedure Install_ASIS is
begin
   Put_Line ("Hello, I am an ASIS Application!");
end Install_ASIS;
