--  This is the test unit to try out the ASIS applications built
--  in the framework of the ASIS tutorial.

package Test_Unit is

   generic
      type T is private;
   procedure Swap (A, B : in out T);

end Test_Unit;