------------------------------------------------------------------------------
--                                                                          --
--                            GNATPP COMPONENTS                             --
--                                                                          --
--                              G N A T P P                                 --
--                                                                          --
--                                 S p e c                                  --
--                                                                          --
--                            $Revision: 207262 $
--                                                                          --
--                     Copyright (C) 2001-2013, AdaCore                     --
--                                                                          --
-- GNATPP is free software; you can redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNATPP is  distributed in the  hope that it will  be  useful, but --
-- WITHOUT ANY WARRANTY; without even the implied warranty of  MERCHANTABI- --
-- LITY or  FITNESS  FOR A  PARTICULAR  PURPOSE. See the GNU General Public --
-- License  for more details. You  should  have  received a copy of the GNU --
-- General Public License  distributed with GNAT; see file COPYING. If not, --
-- write  to  the Free  Software  Foundation,  59 Temple Place - Suite 330, --
-- Boston,                                                                  --
--                                                                          --
-- GNATPP is maintained by ACT Europe (http://www.act-europe.fr).           --
--                                                                          --
------------------------------------------------------------------------------

--  The top of GNATPP hierarchy
--
--  GNATPP is an ASIS-based GNAT pretty-printer

package GNATPP is
   pragma Pure (GNATPP);

   ASIS_2005_Mode : constant Boolean := True;
   --  If this flag is ON, ASIS is initialized in Ada 2005 mode, and the tool
   --  may perform some Ada-2005-specific analysis. ???This flag is always
   --  True, and should be removed.
end GNATPP;
